## usage: [x,rhist]= cg (A,b,x0,Q,options)
##
##    Solve the linear system A*x=b iteratively by the method of Conjugate 
##    Gradients for symmetrx, positive definite matrices. Complex
##    hermitian matrices are allowed.
##
##    A (matrix/string)      : The matrix of coefficients of the linear
##                                 system. 
##                                 If it is a string, say "atimesx" 
##                                 then it should be
##                                 the name of an octave function returning A*x
##                                 when passed x as an argument.
##                                 atimesx(x) should return A*x.
##    b (vector/matrix)      : right hand side. If it is a matrix, then
##                             it represents multiple right hand sides,
##                             and x returns a matrix containing the
##                             corresponding solutions. rhist then
##                             returns the history of convergences for
##                             each r.h.s.
##    x0 (vector, optional)  : initial guess for x
##    Q (matrix/string)      : The preconditioning matrix, i.e. Q*x
##                              should be close to A \ x if you want to
##                              improve convergence.
##                              If it is a string say "preco" then
##                              the octave function preco(x) should
##                              return Q*x. 
##    options ('param1=val1; param2=val2;')
##                           : defines parameters like 
##                               tol=threshold value of residual to converge (default: 1e-10)
##                               nitmax=max. number of iterations (default: Inf)
##                               true_res=(0/1) compute true residuals? (default: 0)
##    x (vector)             : solution of the linear system 
##                                 to prescribed tolerance
##    rhist (vector)         : history of convergence. rhist(k) returns the
##                                 residual at iteration k.

function [x,rhist]= cg(rname_,b,x0,qname_,options)

  if nargin<=1 | nargin >5
    usage('[x,rhist]= cg (''atimesx'',b,x0,Q,options)');
    return
  end 

  if nargin<=4
    options='';
  end

  tol=1e-10;
  nitmax=Inf;
  true_res=0;
  eval(options);

  if nargin<=2
    x0=zeros(size(b));
  end 

  if any(size(x0)~=size(b))
    error('x0 and b should be of the same size')
  end 

  iter=0;
  x=x0;
  if isstr(rname_)
    r=-feval(rname_,x)+b;
  else
    r=-rname_*x+b;
  end 

  rhist=[];
  while 1
    if ~exist('qname_')
      z=r;
    elseif isstr(qname_)
      z=feval(qname_,r);
    else
      z=qname_*r;
    end 
    iter=iter+1;
    rz=sum(conj(r).*z);
    if iter==1
      p=z;
    else
      beta=rz./rzprev;
      p=z+rightscal(p,beta);
    end 
    if isstr(rname_)
      Ap=feval(rname_,p);
    else
      Ap=rname_*p;
    end 
    alpha=rz./sum(conj(p).*Ap);
    x=x+rightscal(p,alpha);
    if true_res
      if isstr(rname_)
        r=-feval(rname_,x)+b;
      else
        r=rname_*x+b;
      end
    else
      r=r-rightscal(Ap,alpha);
    end
    ar=sqrt(real(sum(conj(r).*r)));
    rhist=[rhist;
           ar];
    if (max(ar)<tol | iter>nitmax)
      break;
    end
    rzprev=rz;
  end
endfunction
